<?php

namespace App\Http\Controllers\Admin;

use App\Http\Controllers\Controller;
use App\Models\Category;
use Illuminate\Http\Request;
use Illuminate\Support\Str;
use Illuminate\Support\Facades\File; 
use Validator, Auth, DB, Hash, Image;

class CategoryController extends Controller
{
    public function index()
    {
        $categoryData = Category::all();
        $categoryPageVars = [
            //This is the title of my custom view.
            'pageTitle'=>'Category List',
            'categoryData'=>$categoryData
        ];
        return view('admin.pages.categories.category-list')->with($categoryPageVars);
    }

    public function store(Request $request)
    {
        $data = $request->all();
        $rules = [
            'category_name' => ['required', 'string', 'min:3', 'max:50'],
            'category_image' => ['required', 'image', 'mimes:jpeg,png,jpg,webp', 'max:2048'],
            'status' => ['required']
        ];

        $message = [
            'category_name.required' => 'Category name field is required.',
            'category_image.required' => 'Category image field is required.',
            'category_image.image' => 'Category image upload only image format.',
            'category_image.mimes' => 'Category image upload valid extension only jpeg,png,jpg,webp.',
            'category_image.max' => 'Category image upload size only 2M.',
            'status.required' => 'Status field is required.'
        ];

        $validator = Validator::make($data, $rules, $message);
        if($validator->fails()) {
            return redirect('admin/category')
                        ->withErrors($validator)
                        ->withInput();
        }

        $image = time().'.'.$request->category_image->extension();
        $request->category_image->move(public_path('category'), $image);

        $categoryStore = Category::create([
            "category_name" => $data['category_name'],
            "category_image" => $image,
            "status" => $data['status']
        ]); 

        return redirect('admin/category/displaycategory/')->with('success','Category added successfully.');
    }

    public function show($id){
        $category = Category::find($id);
        return view('admin.pages.categories.edit-category', compact('category'));
    }

    public function update(Request $request)
    {
        $data = $request->all();
        $catId = $data['cat_id'];
        $rules = [
            'category_name' => ['required', 'string', 'min:3', 'max:50'],
            'category_image' => ['image', 'mimes:jpeg,png,jpg,webp', 'max:10240'],
            'status' => ['required']
        ];

        $message = [
            'category_name.required' => 'Category name field is required.',
            'category_image.required' => 'Category image field is required.',
            'category_image.image' => 'Category image upload only image format.',
            'category_image.mimes' => 'Category image upload valid extension only jpeg,png,jpg,webp.',
            'category_image.max' => 'Category image upload size only 5M.',
            'status.required' => 'Status field is required.'
        ];
        $validator = Validator::make($data, $rules, $message);
        if($validator->fails()) {
            return redirect('admin/category/'.$catId)
                        ->withErrors($validator)
                        ->withInput();
        }

        $categorygetdata = Category::select('category_name', 'category_image')->where('id', "=", $catId)->get();
        $singlecategoryname = $categorygetdata[0]->category_name;
        if($singlecategoryname != $request->category_name){
            $slugname = $this->generateSlug($request->category_name);
            $result = Category::where('id', "=", $catId)
                    ->update(['slug' => $slugname]);
        }

        if($request->category_image){
            $categoryimageName = $categorygetdata[0]->category_image;
            File::delete('category/'.$categoryimageName);

            $image = $request->file('category_image');
            $imageName = time().'.'.$request->category_image->extension();
            $destinationPath = public_path('/category');

            $img = Image::make($image->path());
            $img->resize(300, 188, function ($constraint) {
                $constraint->aspectRatio();
            })->save($destinationPath.'/'.$imageName);
            
            // $image->move($destinationPath, $imageName);
            // $request->category_image->move(public_path('category'), $image);

            $result = Category::where('id', "=", $catId)
                    ->update(['category_name' => $data['category_name'], 'category_image' => $imageName, 'status' => $data['status']]);
        }else{
            $result = Category::where('id', "=", $catId)
                    ->update(['category_name' => $data['category_name'], 'status' => $data['status']]);
        }

        if($result){
            return redirect('admin/category/displaycategory/')->with('success','Category updated successfully.');
        }else{
            return redirect('admin/category/'.$catId)->with('error','Category data is not updated!');
        }
    }

    public function delete($id){
        $categorygetdata = Category::select('category_image')->where('id', "=", $id)->get();
        $categoryimageName = $categorygetdata[0]->category_image;
        File::delete('category/'.$categoryimageName);

        $result = Category::where('id', '=', $id)->delete();

        if($result){
            return redirect('admin/category/displaycategory')->with('success','Category record deleted successfully!');
        }
    }

    private function generateSlug($name)
    {
        if (Category::whereSlug($slug = Str::slug($name))->exists()) {
            $max = Category::where("category_name", "=", $name)->latest('id')->skip(1)->value('slug');
            if (isset($max[-1]) && is_numeric($max[-1])) {
                return preg_replace_callback('/(\d+)$/', function($mathces) {
                    return $mathces[1] + 1;
                }, $max);
            }
            return "{$slug}-2";
        }
        return $slug;
    }
}
